// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  let ie_lt8 =
    /MSIE \d/.test(navigator.userAgent) &&
    (document.documentMode == null || document.documentMode < 8)

  let Pos = CodeMirror.Pos

  let matching = {
    '(': ')>',
    ')': '(<',
    '[': ']>',
    ']': '[<',
    '{': '}>',
    '}': '{<',
    '<': '>>',
    '>': '<<',
  }

  function bracketRegex(config) {
    return (config && config.bracketRegex) || /[(){}[\]]/
  }

  function findMatchingBracket(cm, where, config) {
    let line = cm.getLineHandle(where.line),
      pos = where.ch - 1
    let afterCursor = config && config.afterCursor
    if (afterCursor == null)
      afterCursor = /(^| )cm-fat-cursor($| )/.test(
        cm.getWrapperElement().className,
      )
    let re = bracketRegex(config)

    // A cursor is defined as between two characters, but in vim command mode
    // (i.e. not insert mode), the cursor is visually represented as a
    // highlighted box on top of the 2nd character. Otherwise, we allow matches
    // from before or after the cursor.
    let match =
      (!afterCursor &&
        pos >= 0 &&
        re.test(line.text.charAt(pos)) &&
        matching[line.text.charAt(pos)]) ||
      (re.test(line.text.charAt(pos + 1)) && matching[line.text.charAt(++pos)])
    if (!match) return null
    let dir = match.charAt(1) == '>' ? 1 : -1
    if (config && config.strict && dir > 0 != (pos == where.ch)) return null
    let style = cm.getTokenTypeAt(Pos(where.line, pos + 1))

    let found = scanForBracket(
      cm,
      Pos(where.line, pos + (dir > 0 ? 1 : 0)),
      dir,
      style,
      config,
    )
    if (found == null) return null
    return {
      from: Pos(where.line, pos),
      to: found && found.pos,
      match: found && found.ch == match.charAt(0),
      forward: dir > 0,
    }
  }

  // bracketRegex is used to specify which type of bracket to scan
  // should be a regexp, e.g. /[[\]]/
  //
  // Note: If "where" is on an open bracket, then this bracket is ignored.
  //
  // Returns false when no bracket was found, null when it reached
  // maxScanLines and gave up
  function scanForBracket(cm, where, dir, style, config) {
    let maxScanLen = (config && config.maxScanLineLength) || 10000;
    let maxScanLines = (config && config.maxScanLines) || 1000;

    let stack = [];
    let re = bracketRegex(config);
    let lineEnd = getLineEnd(cm, where, dir, maxScanLines);

    for (let lineNo = where.line; (dir > 0 ? lineNo < lineEnd : lineNo > lineEnd); lineNo += dir) {
      let line = cm.getLine(lineNo);
      if (!line || line.length > maxScanLen) continue;

      let pos = getInitialPos(lineNo, where, dir, line.length);
      let end = getEndPos(dir, line.length);

      let result = scanLine(cm, { line, pos, end, dir, re, style, stack, lineNo });
      if (result) return result;
    }
    let boundaryLine = dir > 0 ? cm.lastLine() : cm.firstLine();
    return (lineNo - dir == boundaryLine) ? null : null;  // 始終返回 null
  }


  function getLineEnd(cm, where, dir, maxScanLines) {
    return dir > 0
      ? Math.min(where.line + maxScanLines, cm.lastLine() + 1)
      : Math.max(cm.firstLine() - 1, where.line - maxScanLines);
  }

  function getInitialPos(lineNo, where, dir, lineLength) {
    let adjustment = dir < 0 ? 1 : 0;
    let boundaryPos = dir > 0 ? 0 : lineLength - 1;
    return lineNo == where.line ? where.ch - adjustment : boundaryPos;
  }


  function getEndPos(dir, lineLength) {
    return dir > 0 ? lineLength : -1;
  }

  function scanLine(cm, { line, pos, end, dir, re, style, stack, lineNo }) {
    for (; dir > 0 ? pos < end : pos > end; pos += dir) {
      let ch = line.charAt(pos);
      if (isMatchingBracket(cm, ch, re, style, lineNo, pos)) {
        let match = matching[ch];
        if (match && (match.charAt(1) == '>') == dir > 0) stack.push(ch);
        else if (!stack.length) return { pos: Pos(lineNo, pos), ch: ch };
        else stack.pop();
      }
    }
    return null;
  }


  function isMatchingBracket(cm, ch, re, style, lineNo, pos) {
    return re.test(ch) && (style === undefined || (cm.getTokenTypeAt(Pos(lineNo, pos + 1)) || '') == (style || ''));
  }

  function matchBrackets(cm, autoclear, config) {
    let maxHighlightLen = cm.state.matchBrackets.maxHighlightLineLength || 1000;
    let highlightNonMatching = config && config.highlightNonMatching;
    let marks = [];
    let ranges = cm.listSelections();

    ranges.forEach(range => {
      if (range.empty()) {
        let match = findMatchingBracket(cm, range.head, config);
        if (shouldHighlightMatch(match, highlightNonMatching, cm, maxHighlightLen)) {
          addMatchMarks(cm, match, marks, maxHighlightLen);
        }
      }
    });

    if (marks.length) {
      handleMarks(cm, marks, autoclear);
    }
  }

  function shouldHighlightMatch(match, highlightNonMatching, cm, maxHighlightLen) {
    return match && (match.match || highlightNonMatching !== false) && cm.getLine(match.from.line).length <= maxHighlightLen;
  }

  function addMatchMarks(cm, match, marks, maxHighlightLen) {
    let style = match.match ? 'CodeMirror-matchingbracket' : 'CodeMirror-nonmatchingbracket';
    marks.push(cm.markText(match.from, Pos(match.from.line, match.from.ch + 1), { className: style }));
    if (match.to && cm.getLine(match.to.line).length <= maxHighlightLen) {
      marks.push(cm.markText(match.to, Pos(match.to.line, match.to.ch + 1), { className: style }));
    }
  }

  function handleMarks(cm, marks, autoclear) {
    if (ie_lt8 && cm.state.focused) cm.focus();

    let clear = function () {
      cm.operation(function () {
        marks.forEach(mark => mark.clear());
      });
    };

    if (autoclear) setTimeout(clear, 800);
    else return clear;
  }

  function doMatchBrackets(cm) {
    cm.operation(function () {
      if (cm.state.matchBrackets.currentlyHighlighted) {
        cm.state.matchBrackets.currentlyHighlighted()
        cm.state.matchBrackets.currentlyHighlighted = null
      }
      matchBrackets(
        cm,
        false,
        cm.state.matchBrackets,
      )
    })
  }

  function clearHighlighted(cm) {
    if (cm.state.matchBrackets && cm.state.matchBrackets.currentlyHighlighted) {
      cm.state.matchBrackets.currentlyHighlighted()
      cm.state.matchBrackets.currentlyHighlighted = null
    }
  }

  CodeMirror.defineOption('matchBrackets', false, function (cm, val, old) {
    if (old && old != CodeMirror.Init) {
      cm.off('cursorActivity', doMatchBrackets)
      cm.off('focus', doMatchBrackets)
      cm.off('blur', clearHighlighted)
      clearHighlighted(cm)
    }
    if (val) {
      cm.state.matchBrackets = typeof val == 'object' ? val : {}
      cm.on('cursorActivity', doMatchBrackets)
      cm.on('focus', doMatchBrackets)
      cm.on('blur', clearHighlighted)
    }
  })

  CodeMirror.defineExtension('matchBrackets', function () {
    matchBrackets(this, true)
  })
  CodeMirror.defineExtension(
    'findMatchingBracket',
    function (pos, config, oldConfig) {
      // Backwards-compatibility kludge
      if (oldConfig || typeof config == 'boolean') {
        if (!oldConfig) {
          config = config ? { strict: true } : null
        } else {
          oldConfig.strict = config
          config = oldConfig
        }
      }
      return findMatchingBracket(this, pos, config)
    },
  )
  CodeMirror.defineExtension(
    'scanForBracket',
    function (pos, dir, style, config) {
      let result = scanForBracket(this, pos, dir, style, config);
      return result !== null ? result : null;  // 保證返回類型一致
    }

  )
})
