// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  function bracketFolding(pairs) {
    return function (cm, start) {
      const line = start.line;
      const lineText = cm.getLine(line);

      function findOpening(pair) {
        let tokenType;
        let at = start.ch;
        let pass = 0;

        while (true) {
          const found = at <= 0 ? -1 : lineText.lastIndexOf(pair[0], at - 1);
          if (found === -1) {
            if (pass === 1) break;
            pass = 1;
            at = lineText.length;
            continue;
          }

          if (pass === 1 && found < start.ch) break;
          tokenType = cm.getTokenTypeAt(CodeMirror.Pos(line, found + 1));
          if (!/^(comment|string)/.test(tokenType)) {
            return { ch: found + 1, tokenType, pair };
          }
          at = found - 1;
        }
      }

      function findRange(found) {
        let count = 1;
        let end;
        let startCh = found.ch;
        let endCh;

        for (let i = line; i <= cm.lastLine(); ++i) {
          const text = cm.getLine(i);
          let pos = (i === line) ? startCh : 0;

          while (pos < text.length) {
            const nextOpen = text.indexOf(found.pair[0], pos);
            const nextClose = text.indexOf(found.pair[1], pos);
            pos = Math.min(nextOpen < 0 ? text.length : nextOpen, nextClose < 0 ? text.length : nextClose);

            if (pos === text.length) break;

            const tokenType = cm.getTokenTypeAt(CodeMirror.Pos(i, pos + 1));
            if (tokenType === found.tokenType) {
              if (pos === nextOpen) ++count;
              else if (--count === 0) {
                end = i;
                endCh = pos;
                return { from: CodeMirror.Pos(line, startCh), to: CodeMirror.Pos(end, endCh) };
              }
            }
            ++pos;
          }
        }

        return null;
      }

      // Find all openings
      const found = pairs.map(pair => findOpening(pair)).filter(Boolean);

      // Sort found positions by their character index
      found.sort((a, b) => a.ch - b.ch);

      // Find and return the first matching range
      for (const item of found) {
        const range = findRange(item);
        if (range) return range;
      }


      return null;
    };
  }


  CodeMirror.registerHelper(
    'fold',
    'brace',
    bracketFolding([
      ['{', '}'],
      ['[', ']'],
    ]),
  )

  CodeMirror.registerHelper(
    'fold',
    'brace-paren',
    bracketFolding([
      ['{', '}'],
      ['[', ']'],
      ['(', ')'],
    ]),
  )

  CodeMirror.registerHelper('fold', 'import', function (cm, start) {
    function getImportStart(line) {
      if (line < cm.firstLine() || line > cm.lastLine()) return null;
      let start = cm.getTokenAt(CodeMirror.Pos(line, 1));
      if (!/\S/.test(start.string)) start = cm.getTokenAt(CodeMirror.Pos(line, start.end + 1));
      return (start.type === 'keyword' && start.string === 'import') ? start : null;
    }

    function findClosingSemicolon(line) {
      for (let i = line, e = Math.min(cm.lastLine(), line + 10); i <= e; ++i) {
        let text = cm.getLine(i);
        let semi = text.indexOf(';');
        if (semi !== -1) return CodeMirror.Pos(i, semi);
      }
      return null;
    }

    function isImportOnPreviousLine(line) {
      const importOnCurrentLine = hasImport(line);
      const importOnPreviousLine = hasImport(line - 1);
      const importOnSecondPreviousLine = hasImport(line - 2) && hasImport(line - 2).end.line === line - 1;

      return importOnCurrentLine || importOnPreviousLine || importOnSecondPreviousLine;
    }


    function hasImport(line) {
      let start = getImportStart(line);
      if (!start) return null;
      let semicolonPos = findClosingSemicolon(line);
      return semicolonPos ? { startCh: start.end, end: semicolonPos } : null;
    }

    let startLine = start.line;
    let importInfo = hasImport(startLine);

    if (!importInfo || isImportOnPreviousLine(startLine)) return null;

    let end = importInfo.end;
    while (true) {
      let nextImport = hasImport(end.line + 1);
      if (!nextImport) break;
      end = nextImport.end;
    }

    return {
      from: cm.clipPos(CodeMirror.Pos(startLine, importInfo.startCh + 1)),
      to: end,
    };
  });


  CodeMirror.registerHelper('fold', 'include', function (cm, start) {
    function hasInclude(line) {
      if (line < cm.firstLine() || line > cm.lastLine()) return null
      let start = cm.getTokenAt(CodeMirror.Pos(line, 1))
      if (!/\S/.test(start.string))
        start = cm.getTokenAt(CodeMirror.Pos(line, start.end + 1))
      if (start.type == 'meta' && start.string.slice(0, 8) == '#include')
        return start.start + 8
    }

    let startLine = start.line,
      has = hasInclude(startLine)
    if (has == null || hasInclude(startLine - 1) != null) return null
    for (let end = startLine; ;) {
      let next = hasInclude(end + 1)
      if (next == null) break
      ++end
    }
    return {
      from: CodeMirror.Pos(startLine, has + 1),
      to: cm.clipPos(CodeMirror.Pos(end)),
    }
  })
})
