// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  let Pos = CodeMirror.Pos
  function cmp(a, b) {
    return a.line - b.line || a.ch - b.ch
  }

  let nameStartChar =
    'A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD';
  let nameChar = nameStartChar + '-:.0-9\\u00B7\\u0300\\u036F\\u203F-\\u2040';
  let xmlTagStart = new RegExp(
    '<(/?)([' + nameStartChar + '][' + nameChar + ']*)',
    'g',
  );


  function Iter(cm, line, ch, range) {
    this.line = line
    this.ch = ch
    this.cm = cm
    this.text = cm.getLine(line)
    this.min = range ? Math.max(range.from, cm.firstLine()) : cm.firstLine()
    this.max = range ? Math.min(range.to - 1, cm.lastLine()) : cm.lastLine()
  }

  function tagAt(iter, ch) {
    let type = iter.cm.getTokenTypeAt(Pos(iter.line, ch))
    return type && /\btag\b/.test(type)
  }

  function nextLine(iter) {
    if (iter.line >= iter.max) return
    iter.ch = 0
    iter.text = iter.cm.getLine(++iter.line)
    return true
  }
  function prevLine(iter) {
    if (iter.line <= iter.min) return
    iter.text = iter.cm.getLine(--iter.line)
    iter.ch = iter.text.length
    return true
  }

  function toTagEnd(iter) {
    for (; ;) {
      let gt = iter.text.indexOf('>', iter.ch)
      if (gt == -1) {
        if (nextLine(iter)) continue
        else return
      }
      if (!tagAt(iter, gt + 1)) {
        iter.ch = gt + 1
        continue
      }
      let lastSlash = iter.text.lastIndexOf('/', gt)
      let selfClose =
        lastSlash > -1 && !/\S/.test(iter.text.slice(lastSlash + 1, gt))
      iter.ch = gt + 1
      return selfClose ? 'selfClose' : 'regular'
    }
  }
  function toTagStart(iter) {
    for (; ;) {
      let lt = iter.ch ? iter.text.lastIndexOf('<', iter.ch - 1) : -1
      if (lt == -1) {
        if (prevLine(iter)) continue
        else return
      }
      if (!tagAt(iter, lt + 1)) {
        iter.ch = lt
        continue
      }
      xmlTagStart.lastIndex = lt
      iter.ch = lt
      let match = xmlTagStart.exec(iter.text)
      if (match && match.index == lt) return match
    }
  }

  function toNextTag(iter) {
    for (; ;) {
      xmlTagStart.lastIndex = iter.ch
      let found = xmlTagStart.exec(iter.text)
      if (!found) {
        if (nextLine(iter)) continue
        else return
      }
      if (!tagAt(iter, found.index + 1)) {
        iter.ch = found.index + 1
        continue
      }
      iter.ch = found.index + found[0].length
      return found
    }
  }
  function toPrevTag(iter) {
    for (; ;) {
      let gt = iter.ch ? iter.text.lastIndexOf('>', iter.ch - 1) : -1
      if (gt == -1) {
        if (prevLine(iter)) continue
        else return
      }
      if (!tagAt(iter, gt + 1)) {
        iter.ch = gt
        continue
      }
      let lastSlash = iter.text.lastIndexOf('/', gt)
      let selfClose =
        lastSlash > -1 && !/\S/.test(iter.text.slice(lastSlash + 1, gt))
      iter.ch = gt + 1
      return selfClose ? 'selfClose' : 'regular'
    }
  }

  function findMatchingClose(iter, tag) {
    let stack = [];

    while (true) {
      let next = toNextTag(iter);
      if (!next) return;  // Early exit if there's no next tag

      let end = toTagEnd(iter);
      if (!end) return;  // Early exit if there's no tag end

      let startLine = iter.line;
      let startCh = iter.ch - (next ? next[0].length : 0);

      if (end === 'selfClose') continue; // Ignore self-closing tags

      if (next[1]) {
        // Process closing tag
        let result = handleClosingTag(stack, next[2], startLine, startCh, iter, tag);
        if (result) return result;
      } else {
        // Process opening tag
        stack.push(next[2]);
      }
    }
  }

  function handleClosingTag(stack, currentTag, startLine, startCh, iter, tag) {
    for (let i = stack.length - 1; i >= 0; --i) {
      if (stack[i] === currentTag) {
        stack.length = i;
        break;
      }
    }

    // If no matching opening tag is found, and the provided tag matches or is not specified
    if (stack.length === 0 && (!tag || tag === currentTag)) {
      return {
        tag: currentTag,
        from: Pos(startLine, startCh),
        to: Pos(iter.line, iter.ch),
      };
    }
    return null;
  }

  function findMatchingOpen(iter, tag) {
    let stack = [];

    while (true) {
      let prev = toPrevTag(iter);
      if (!prev) return;  // Early return if no previous tag

      if (prev === 'selfClose') {
        toTagStart(iter);
        continue;
      }

      let { startLine, startCh, startTag } = handleTagStart(iter);
      if (!startTag) return;

      if (start[1]) {
        // Closing tag
        stack.push(startTag);
      } else {
        // Opening tag, check for matching closing
        let result = handleOpeningTag(stack, startTag, iter, tag, startLine, startCh);
        if (result) return result;
      }
    }
  }

  function handleTagStart(iter) {
    const start = toTagStart(iter);
    if (!start) return { startTag: null };

    const startLine = iter.line;
    const startCh = iter.ch;
    const startTag = start[2];

    return { startLine, startCh, startTag };
  }

  function handleOpeningTag(stack, currentTag, iter, tag, startLine, startCh) {
    // Find matching opening tag by checking the stack
    for (let i = stack.length - 1; i >= 0; --i) {
      if (stack[i] === currentTag) {
        stack.length = i;
        break;
      }
    }

    if (stack.length === 0 && (!tag || tag === currentTag)) {
      return {
        tag: currentTag,
        from: Pos(iter.line, iter.ch),
        to: Pos(startLine, startCh),
      };
    }
    return null;
  }


  CodeMirror.registerHelper('fold', 'xml', function (cm, start) {
    let iter = new Iter(cm, start.line, 0)
    for (; ;) {
      let openTag = toNextTag(iter)
      if (!openTag || iter.line != start.line) return
      let end = toTagEnd(iter)
      if (!end) return
      if (!openTag[1] && end != 'selfClose') {
        let startPos = Pos(iter.line, iter.ch)
        let endPos = findMatchingClose(iter, openTag[2])
        return endPos && cmp(endPos.from, startPos) > 0
          ? { from: startPos, to: endPos.from }
          : null
      }
    }
  })
  CodeMirror.findMatchingTag = function (cm, pos, range) {
    let iter = new Iter(cm, pos.line, pos.ch, range)
    if (iter.text.indexOf('>') == -1 && iter.text.indexOf('<') == -1) return
    let end = toTagEnd(iter),
      to = end && Pos(iter.line, iter.ch)
    let start = end && toTagStart(iter)
    if (!end || !start || cmp(iter, pos) > 0) return
    let here = { from: Pos(iter.line, iter.ch), to: to, tag: start[2] }
    if (end == 'selfClose') return { open: here, close: null, at: 'open' }

    if (start[1]) {
      // closing tag
      return {
        open: findMatchingOpen(iter, start[2]),
        close: here,
        at: 'close',
      }
    } else {
      // opening tag
      iter = new Iter(cm, to.line, to.ch, range)
      return {
        open: here,
        close: findMatchingClose(iter, start[2]),
        at: 'open',
      }
    }
  }

  CodeMirror.findEnclosingTag = function (cm, pos, range, tag) {
    let iter = new Iter(cm, pos.line, pos.ch, range)
    for (; ;) {
      let open = findMatchingOpen(iter, tag)
      if (!open) break
      let forward = new Iter(cm, pos.line, pos.ch, range)
      let close = findMatchingClose(forward, open.tag)
      if (close) return { open: open, close: close }
    }
  }

  // Used by addon/edit/closetag.js
  CodeMirror.scanForClosingTag = function (cm, pos, name, end) {
    let iter = new Iter(cm, pos.line, pos.ch, end ? { from: 0, to: end } : null)
    return findMatchingClose(iter, name)
  }
})
