// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  CodeMirror.defineSimpleMode = function (name, states) {
    CodeMirror.defineMode(name, function (config) {
      return CodeMirror.simpleMode(config, states)
    })
  }

  CodeMirror.simpleMode = function (config, states) {
    ensureState(states, 'start');
    let { states_, hasIndentation } = processStates(states);
    let mode = createMode(states_, config, hasIndentation, states.meta || {});
    return mode;
  }

  function processStates(states) {
    let states_ = {};
    let hasIndentation = false;
    for (let state in states) {
      if (state != 'meta' && states.hasOwnProperty(state)) {
        let list = (states_[state] = []);
        let orig = states[state];
        for (let data of orig) {
          list.push(new Rule(data, states));
          if (data.indent || data.dedent) hasIndentation = true;
        }
      }
    }
    return { states_, hasIndentation };
  }

  function createMode(states_, config, hasIndentation, meta) {
    let mode = {
      startState: function () {
        return {
          state: 'start',
          pending: null,
          local: null,
          localState: null,
          indent: hasIndentation ? [] : null,
        };
      },
      copyState: function (state) {
        return copyState(state, hasIndentation);
      },
      token: tokenFunction(states_, config),
      innerMode: function (state) {
        return state.local && { mode: state.local.mode, state: state.localState };
      },
      indent: indentFunction(states_, meta),
    };
    if (meta) {
      for (let prop in meta) {
        if (meta.hasOwnProperty(prop)) mode[prop] = meta[prop];
      }
    }
    return mode;
  }

  function copyState(state, hasIndentation) {
    let s = {
      state: state.state,
      pending: state.pending,
      local: state.local,
      localState: null,
      indent: state.indent && state.indent.slice(0),
    };
    if (state.localState)
      s.localState = CodeMirror.copyState(state.local.mode, state.localState);
    if (state.stack) s.stack = state.stack.slice(0);
    for (let pers = state.persistentStates; pers; pers = pers.next)
      s.persistentStates = {
        mode: pers.mode,
        spec: pers.spec,
        state: pers.state == state.localState ? s.localState : CodeMirror.copyState(pers.mode, pers.state),
        next: s.persistentStates,
      };
    return s;
  }

  function ensureState(states, name) {
    if (!states.hasOwnProperty(name))
      throw new Error('Undefined state ' + name + ' in simple mode')
  }

  function toRegex(val, caret) {
    if (!val) return /./
    let flags = ''
    if (val instanceof RegExp) {
      if (val.ignoreCase) flags = 'i'
      if (val.unicode) flags += 'u'
      val = val.source
    } else {
      val = String(val)
    }
    return new RegExp((caret === false ? '' : '^') + '(?:' + val + ')', flags)
  }

  function asToken(val) {
    if (!val) return null
    if (val.apply) return val
    if (typeof val == 'string') return val.replace(/\./g, ' ')
    let result = []
    for (let item of val) {
      result.push(item && item.replace(/\./g, ' '));
    }
    return result
  }

  function Rule(data, states) {
    if (data.next || data.push) ensureState(states, data.next || data.push)
    this.regex = toRegex(data.regex)
    this.token = asToken(data.token)
    this.data = data
  }

  function tokenFunction(states, config) {
    return function (stream, state) {
      if (state.pending) {
        return handlePending(stream, state);
      }

      if (state.local) {
        return handleLocal(stream, state);
      }

      return handleState(stream, state, states, config);
    }
  }

  function handlePending(stream, state) {
    let pend = state.pending.shift();
    if (state.pending.length == 0) state.pending = null;
    stream.pos += pend.text.length;
    return pend.token;
  }

  function handleLocal(stream, state) {
    if (state.local.end && stream.match(state.local.end)) {
      let tok = state.local.endToken || null;
      state.local = state.localState = null;
      return tok;
    } else {
      let tok = state.local.mode.token(stream, state.localState),
        m;
      if (state.local.endScan && (m = state.local.endScan.exec(stream.current()))) {
        stream.pos = stream.start + m.index;
      }
      return tok;
    }
  }

  function handleState(stream, state, states, config) {
    let curState = states[state.state];
    for (let rule of curState) {
      let matches = (!rule.data.sol || stream.sol()) && stream.match(rule.regex);
      if (matches) {
        updateState(state, rule, config, stream, matches);
        return getToken(rule, matches, state, stream);
      }
    }
    stream.next();
    return null;
  }

  function updateState(state, rule, config, stream, matches) {
    if (rule.data.next) {
      state.state = rule.data.next;
    } else if (rule.data.push) {
      (state.stack || (state.stack = [])).push(state.state);
      state.state = rule.data.push;
    } else if (rule.data.pop && state.stack && state.stack.length) {
      state.state = state.stack.pop();
    }

    if (rule.data.mode) enterLocalMode(config, state, rule.data.mode, rule.token);
    if (rule.data.indent) state.indent.push(stream.indentation() + config.indentUnit);
    if (rule.data.dedent) state.indent.pop();
  }

  function getToken(rule, matches, state, stream) {
    let token = rule.token;
    if (token && token.apply) token = token(matches);
    if (matches.length > 2 && rule.token && typeof rule.token != 'string') {
      for (let j = 2; j < matches.length; j++)
        if (matches[j])
          (state.pending || (state.pending = [])).push({
            text: matches[j],
            token: rule.token[j - 1],
          });
      stream.backUp(matches[0].length - (matches[1] ? matches[1].length : 0));
      return token[0];
    } else if (token && token.join) {
      return token[0];
    } else {
      return token;
    }
  }

  function cmp(a, b) {
    if (a === b) return true
    if (!a || typeof a != 'object' || !b || typeof b != 'object') return false
    let props = 0
    for (let prop in a)
      if (a.hasOwnProperty(prop)) {
        if (!b.hasOwnProperty(prop) || !cmp(a[prop], b[prop])) return false
        props++
      }
    for (let prop in b) if (b.hasOwnProperty(prop)) props--
    return props == 0
  }

  function enterLocalMode(config, state, spec, token) {
    let pers = findPersistentState(state, spec);
    let mode = pers ? pers.mode : spec.mode || CodeMirror.getMode(config, spec.spec);
    let lState = pers ? pers.state : CodeMirror.startState(mode);

    if (spec.persistent && !pers) {
      state.persistentStates = {
        mode: mode,
        spec: spec.spec,
        state: lState,
        next: state.persistentStates,
      };
    }

    state.localState = lState;
    state.local = createLocalState(mode, spec, token);
  }

  function findPersistentState(state, spec) {
    for (let p = state.persistentStates; p; p = p.next) {
      if (spec.spec ? cmp(spec.spec, p.spec) : spec.mode == p.mode) {
        return p;
      }
    }
    return null;
  }

  function createLocalState(mode, spec, token) {
    return {
      mode: mode,
      end: spec.end && toRegex(spec.end),
      endScan: spec.end && spec.forceEnd !== false && toRegex(spec.end, false),
      endToken: token && token.join ? token[token.length - 1] : token,
    };
  }

  function indexOf(val, arr) {
    for (let item of arr) {
      if (item === val) return true;
    }
  }

  function indentFunction(states, meta) {
    return function (state, textAfter, line) {
      if (shouldPassIndent(state, meta)) {
        return CodeMirror.Pass;
      }

      if (state.local && state.local.mode.indent) {
        return state.local.mode.indent(state.localState, textAfter, line);
      }

      let pos = state.indent.length - 1;
      let rules = states[state.state];
      pos = processRules(rules, textAfter, pos, states);

      return pos < 0 ? 0 : state.indent[pos];
    }
  }

  function shouldPassIndent(state, meta) {
    return state.indent == null ||
      state.local ||
      (meta.dontIndentStates && indexOf(state.state, meta.dontIndentStates) > -1);
  }

  function processRules(rules, textAfter, pos, states) {
    let continueScan = true;

    while (continueScan) {
      continueScan = false;
      for (let rule of rules) {
        if (shouldProcessRule(rule)) {
          let m = rule.regex.exec(textAfter);
          if (m && m[0]) {
            pos = updatePositionAndText(pos, m[0], textAfter, rule, states);
            continueScan = true;
            break;
          }
        }
      }
    }

    return pos;
  }

  function shouldProcessRule(rule) {
    return rule.data.dedent && rule.data.dedentIfLineStart !== false;
  }

  function updatePositionAndText(pos, matchedText, textAfter, rule, states) {
    pos--;  // Adjust the position after matching
    if (rule.next || rule.push) {
      rules = states[rule.next || rule.push];  // Update rules based on next or push state
    }
    textAfter = textAfter.slice(matchedText.length);  // Remove matched text from the beginning
    return pos;
  }

})
