// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  CodeMirror.defineExtension('annotateScrollbar', function (options) {
    if (typeof options == 'string') options = { className: options }
    return new Annotation(this, options)
  })

  CodeMirror.defineOption('scrollButtonHeight', 0)

  function Annotation(cm, options) {
    this.cm = cm
    this.options = options
    this.buttonHeight =
      options.scrollButtonHeight || cm.getOption('scrollButtonHeight')
    this.annotations = []
    this.doRedraw = this.doUpdate = null
    this.div = cm.getWrapperElement().appendChild(document.createElement('div'))
    this.div.style.cssText =
      'position: absolute; right: 0; top: 0; z-index: 7; pointer-events: none'
    this.computeScale()

    function scheduleRedraw(delay) {
      clearTimeout(self.doRedraw)
      self.doRedraw = setTimeout(function () {
        self.redraw()
      }, delay)
    }

    let self = this
    this.resizeHandler = function () {
      clearTimeout(self.doUpdate)
      self.doUpdate = setTimeout(function () {
        if (self.computeScale()) scheduleRedraw(20)
      }, 100)
    }
    cm.on('refresh', this.resizeHandler)
    cm.on('markerAdded', this.resizeHandler)
    cm.on('markerCleared', this.resizeHandler)
    if (options.listenForChanges !== false) {
      this.changeHandler = function () {
        scheduleRedraw(250)
      }
      cm.on('changes', this.changeHandler)
    }
  }

  Annotation.prototype.computeScale = function () {
    let cm = this.cm
    let hScale =
      (cm.getWrapperElement().clientHeight -
        cm.display.barHeight -
        this.buttonHeight * 2) /
      cm.getScrollerElement().scrollHeight
    if (hScale != this.hScale) {
      this.hScale = hScale
      return true
    }
  }

  Annotation.prototype.update = function (annotations) {
    this.annotations = annotations
    this.redraw()
  }

  Annotation.prototype.redraw = function (compute) {
    if (compute !== false) this.computeScale()
    let cm = this.cm,
      hScale = this.hScale

    let frag = document.createDocumentFragment(),
      anns = this.annotations

    let wrapping = cm.getOption('lineWrapping')
    let singleLineH = wrapping && cm.defaultTextHeight() * 1.5
    let curLine = null,
      curLineObj = null

    const getY = (pos, top) => {
      if (curLine != pos.line) {
        curLine = pos.line
        curLineObj = cm.getLineHandle(pos.line)
        let visual = cm.getLineHandleVisualStart(curLineObj)
        if (visual != curLineObj) {
          curLine = cm.getLineNumber(visual)
          curLineObj = visual
        }
      }
      if (
        (curLineObj.widgets && curLineObj.widgets.length) ||
        (wrapping && curLineObj.height > singleLineH)
      )
        return cm.charCoords(pos, 'local')[top ? 'top' : 'bottom']
      let topY = cm.heightAtLine(curLineObj, 'local')
      return topY + (top ? 0 : curLineObj.height)
    }

    const createAnnotationElement = (top, bottom, ann) => {
      let height = Math.max(bottom - top, 3)
      let elt = document.createElement('div')
      elt.style.cssText =
        'position: absolute; right: 0px; width: ' +
        Math.max(cm.display.barWidth - 1, 2) +
        'px; top: ' +
        (top + this.buttonHeight) +
        'px; height: ' +
        height +
        'px'
      elt.className = this.options.className
      if (ann.id) {
        elt.setAttribute('annotation-id', ann.id)
      }
      return elt
    }

    const processAnnotations = () => {
      let lastLine = cm.lastLine()
      if (cm.display.barWidth) {
        for (let i = 0, nextTop; i < anns.length; i++) {
          let ann = anns[i]
          if (ann.to.line > lastLine) continue
          let top = nextTop || getY(ann.from, true) * hScale
          let bottom = getY(ann.to, false) * hScale
          let i = 0;
          while (i < anns.length - 1) {
            if (anns[i + 1].to.line > lastLine) break;
            nextTop = getY(anns[i + 1].from, true) * hScale;
            if (nextTop > bottom + 0.9) break;
            ann = anns[i + 1];
            bottom = getY(ann.to, false) * hScale;
            i = Math.min(i + 1, anns.length - 1); // 控制 i 不會超過陣列範圍
          }

          if (bottom == top) continue
          let elt = createAnnotationElement(top, bottom, ann)
          frag.appendChild(elt)
        }
      }
    }

    processAnnotations()
    this.div.textContent = ''
    this.div.appendChild(frag)
  }

  Annotation.prototype.clear = function () {
    this.cm.off('refresh', this.resizeHandler)
    this.cm.off('markerAdded', this.resizeHandler)
    this.cm.off('markerCleared', this.resizeHandler)
    if (this.changeHandler) this.cm.off('changes', this.changeHandler)
    this.div.parentNode.removeChild(this.div)
  }
})
