// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(require('../../lib/codemirror'))
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  function Bar(cls, orientation, scroll) {
    this.orientation = orientation
    this.scroll = scroll
    this.screen = this.total = this.size = 1
    this.pos = 0

    this.node = document.createElement('div')
    this.node.className = cls + '-' + orientation
    this.inner = this.node.appendChild(document.createElement('div'))

    CodeMirror.on(this.inner, 'mousedown', onMouseDown.bind(this))
    CodeMirror.on(this.node, 'click', onClick.bind(this))
    CodeMirror.on(this.node, 'mousewheel', onWheel.bind(this))
    CodeMirror.on(this.node, 'DOMMouseScroll', onWheel.bind(this))
  }

  function onMouseDown(e) {
    if (e.which != 1) return
    CodeMirror.e_preventDefault(e)
    let axis = this.orientation == 'horizontal' ? 'pageX' : 'pageY'
    let start = e[axis],
      startpos = this.pos
    let self = this

    function done() {
      CodeMirror.off(document, 'mousemove', move)
      CodeMirror.off(document, 'mouseup', done)
    }

    function move(e) {
      if (e.which != 1) return done()
      self.moveTo(startpos + (e[axis] - start) * (self.total / self.size))
    }

    CodeMirror.on(document, 'mousemove', move)
    CodeMirror.on(document, 'mouseup', done)
  }

  function onClick(e) {
    CodeMirror.e_preventDefault(e)
    let innerBox = this.inner.getBoundingClientRect(),
      where
    if (this.orientation == 'horizontal') {
      if (e.clientX < innerBox.left) {
        where = -1;
      } else if (e.clientX > innerBox.right) {
        where = 1;
      } else {
        where = 0;
      }
    } else {
      if (e.clientY < innerBox.top) {
        where = -1;
      } else if (e.clientY > innerBox.bottom) {
        where = 1;
      } else {
        where = 0;
      }
    }
    this.moveTo(this.pos + where * this.screen)
  }

  function onWheel(e) {
    let moved = CodeMirror.wheelEventPixels(e)[
      this.orientation == 'horizontal' ? 'x' : 'y'
    ]
    let oldPos = this.pos
    this.moveTo(this.pos + moved)
    if (this.pos != oldPos) CodeMirror.e_preventDefault(e)
  }

  Bar.prototype.setPos = function (pos, force) {
    if (pos < 0) pos = 0
    if (pos > this.total - this.screen) pos = this.total - this.screen
    if (!force && pos == this.pos) return false
    this.pos = pos
    this.inner.style[this.orientation == 'horizontal' ? 'left' : 'top'] =
      pos * (this.size / this.total) + 'px'
    return true
  }

  Bar.prototype.moveTo = function (pos) {
    if (this.setPos(pos)) this.scroll(pos, this.orientation)
  }

  let minButtonSize = 10

  Bar.prototype.update = function (scrollSize, clientSize, barSize) {
    let sizeChanged =
      this.screen != clientSize ||
      this.total != scrollSize ||
      this.size != barSize
    if (sizeChanged) {
      this.screen = clientSize
      this.total = scrollSize
      this.size = barSize
    }

    let buttonSize = this.screen * (this.size / this.total)
    if (buttonSize < minButtonSize) {
      this.size -= minButtonSize - buttonSize
      buttonSize = minButtonSize
    }
    this.inner.style[this.orientation == 'horizontal' ? 'width' : 'height'] =
      buttonSize + 'px'
    this.setPos(this.pos, sizeChanged)
  }

  function SimpleScrollbars(cls, place, scroll) {
    this.addClass = cls
    this.horiz = new Bar(cls, 'horizontal', scroll)
    place(this.horiz.node)
    this.vert = new Bar(cls, 'vertical', scroll)
    place(this.vert.node)
    this.width = null
  }

  SimpleScrollbars.prototype.update = function (measure) {
    if (this.width == null) {
      this.width = getScrollbarWidth(this.horiz.node);
    }
    let width = this.width || 0;

    let needsH = measure.scrollWidth > measure.clientWidth + 1;
    let needsV = measure.scrollHeight > measure.clientHeight + 1;
    this.vert.node.style.display = needsV ? 'block' : 'none';
    this.horiz.node.style.display = needsH ? 'block' : 'none';

    if (needsV) {
      updateVerticalScrollbar(this.vert, measure, needsH, width);
    }
    if (needsH) {
      updateHorizontalScrollbar(this.horiz, measure, needsV, width);
    }

    return { right: needsV ? width : 0, bottom: needsH ? width : 0 };
  }

  function getScrollbarWidth(node) {
    let style = window.getComputedStyle ? window.getComputedStyle(node) : node.currentStyle;
    return style ? parseInt(style.height) : 0;
  }

  function updateVerticalScrollbar(vert, measure, needsH, width) {
    vert.update(
      measure.scrollHeight,
      measure.clientHeight,
      measure.viewHeight - (needsH ? width : 0),
    );
    vert.node.style.bottom = needsH ? width + 'px' : '0';
  }

  function updateHorizontalScrollbar(horiz, measure, needsV, width) {
    horiz.update(
      measure.scrollWidth,
      measure.clientWidth,
      measure.viewWidth - (needsV ? width : 0) - measure.barLeft,
    );
    horiz.node.style.right = needsV ? width + 'px' : '0';
    horiz.node.style.left = measure.barLeft + 'px';
  }

  SimpleScrollbars.prototype.setScrollTop = function (pos) {
    this.vert.setPos(pos)
  }

  SimpleScrollbars.prototype.setScrollLeft = function (pos) {
    this.horiz.setPos(pos)
  }

  SimpleScrollbars.prototype.clear = function () {
    let parent = this.horiz.node.parentNode
    parent.removeChild(this.horiz.node)
    parent.removeChild(this.vert.node)
  }

  CodeMirror.scrollbarModel.simple = function (place, scroll) {
    return new SimpleScrollbars('CodeMirror-simplescroll', place, scroll)
  }
  CodeMirror.scrollbarModel.overlay = function (place, scroll) {
    return new SimpleScrollbars('CodeMirror-overlayscroll', place, scroll)
  }
})
