// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

// Define search commands. Depends on dialog.js or another
// implementation of the openDialog method.

// Replace works a little oddly -- it will do the replace on the next
// Ctrl-G (or whatever is bound to findNext) press. You prevent a
// replace by making sure the match is no longer selected when hitting
// Ctrl-G.

; (function (mod) {
  if (typeof exports == 'object' && typeof module == 'object')
    // CommonJS
    mod(
      require('../../lib/codemirror'),
      require('./searchcursor'),
      require('../dialog/dialog'),
    )
  else if (typeof define == 'function' && define.amd)
    // AMD
    define(['../../lib/codemirror', './searchcursor', '../dialog/dialog'], mod)
  // Plain browser env
  else mod(CodeMirror)
})(function (CodeMirror) {
  'use strict'

  // default search panel location
  CodeMirror.defineOption('search', { bottom: false })

  function searchOverlay(query, caseInsensitive) {
    if (typeof query == 'string')
      query = new RegExp(
        query.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&'),
        caseInsensitive ? 'gi' : 'g',
      )
    else if (!query.global)
      query = new RegExp(query.source, query.ignoreCase ? 'gi' : 'g')

    return {
      token: function (stream) {
        query.lastIndex = stream.pos
        let match = query.exec(stream.string)
        if (match && match.index == stream.pos) {
          stream.pos += match[0].length || 1
          return 'searching'
        } else if (match) {
          stream.pos = match.index
        } else {
          stream.skipToEnd()
        }
      },
    }
  }

  function SearchState() {
    this.posFrom = this.posTo = this.lastQuery = this.query = null
    this.overlay = null
  }

  function getSearchState(cm) {
    return cm.state.search || (cm.state.search = new SearchState())
  }

  function queryCaseInsensitive(query) {
    return typeof query == 'string' && query == query.toLowerCase()
  }

  function getSearchCursor(cm, query, pos) {
    // Heuristic: if the query string is all lowercase, do a case insensitive search.
    return cm.getSearchCursor(query, pos, {
      caseFold: queryCaseInsensitive(query),
      multiline: true,
    })
  }

  function persistentDialog(cm, text, deflt, onEnter, onKeyDown) {
    cm.openDialog(text, onEnter, {
      value: deflt,
      selectValueOnOpen: true,
      closeOnEnter: false,
      onClose: function () {
        clearSearch(cm)
      },
      onKeyDown: onKeyDown,
      bottom: cm.options.search.bottom,
    })
  }

  function dialog(cm, text, shortText, deflt, f) {
    if (cm.openDialog)
      cm.openDialog(text, f, {
        value: deflt,
        selectValueOnOpen: true,
        bottom: cm.options.search.bottom,
      })
    else f(prompt(shortText, deflt))
  }

  function confirmDialog(cm, text, shortText, fs) {
    if (cm.openConfirm) cm.openConfirm(text, fs)
    else if (confirm(shortText)) fs[0]()
  }

  function parseString(string) {
    return string.replace(/\\([nrt\\])/g, function (match, ch) {
      if (ch == 'n') return '\n'
      if (ch == 'r') return '\r'
      if (ch == 't') return '\t'
      if (ch == '\\') return '\\'
      return match
    })
  }

  function parseQuery(query) {
    let isRE = query.match(/^\/(.*)\/([a-z]*)$/)
    if (isRE) {
      try {
        query = new RegExp(isRE[1], isRE[2].indexOf('i') == -1 ? '' : 'i')
      } catch (e) { } // Not a regular expression after all, do a string search
    } else {
      query = parseString(query)
    }
    if (typeof query == 'string' ? query == '' : query.test('')) query = /x^/
    return query
  }

  function startSearch(cm, state, query) {
    state.queryText = query
    state.query = parseQuery(query)
    cm.removeOverlay(state.overlay, queryCaseInsensitive(state.query))
    state.overlay = searchOverlay(
      state.query,
      queryCaseInsensitive(state.query),
    )
    cm.addOverlay(state.overlay)
    if (cm.showMatchesOnScrollbar) {
      if (state.annotate) {
        state.annotate.clear()
        state.annotate = null
      }
      state.annotate = cm.showMatchesOnScrollbar(
        state.query,
        queryCaseInsensitive(state.query),
      )
    }
  }

  function doSearch(cm, rev, persistent, immediate) {
    const state = getSearchState(cm);

    // If there is an active query, proceed with finding the next match.
    if (state.query) return findNext(cm, rev);

    let query = cm.getSelection() || state.lastQuery;
    if (query instanceof RegExp && query.source == 'x^') query = null;

    if (persistent && cm.openDialog) {
      handlePersistentSearch(cm, state, query, rev, immediate);
    } else {
      handleNonPersistentSearch(cm, state, query, rev);
    }
  }

  function handlePersistentSearch(cm, state, query, rev, immediate) {
    let hiding = null;

    const searchNext = (query, event) => {
      CodeMirror.e_stop(event);
      if (!query) return;

      if (query != state.queryText) {
        startSearch(cm, state, query);
        state.posFrom = state.posTo = cm.getCursor();
      }

      if (hiding) hiding.style.opacity = 1;
      findNext(cm, event.shiftKey, function (_, to) {
        if (shouldHideDialog(cm, to)) {
          hiding.style.opacity = 0.4;
        }
      });
    };

    persistentDialog(
      cm,
      getQueryDialog(cm),
      query,
      searchNext,
      function (event, query) {
        handleDialogKeyPress(event, query, cm, state);
      }
    );

    if (immediate && query) {
      startSearch(cm, state, query);
      findNext(cm, rev);
    }
  }

  function handleNonPersistentSearch(cm, state, query, rev) {
    dialog(cm, getQueryDialog(cm), 'Search for:', query, function (query) {
      if (query && !state.query) {
        cm.operation(function () {
          startSearch(cm, state, query);
          state.posFrom = state.posTo = cm.getCursor();
          findNext(cm, rev);
        });
      }
    });
  }

  function handleDialogKeyPress(event, query, cm, state) {
    const keyName = CodeMirror.keyName(event);
    const extra = cm.getOption('extraKeys');
    const cmd = (extra && extra[keyName]) || CodeMirror.keyMap[cm.getOption('keyMap')][keyName];

    if (isFindCommand(cmd)) {
      CodeMirror.e_stop(event);
      startSearch(cm, state, query);
      cm.execCommand(cmd);
    } else if (isSearchCommand(cmd)) {
      CodeMirror.e_stop(event);
      searchNext(query, event);
    }
  }

  function shouldHideDialog(cm, to) {
    const dialog = cm.display.wrapper.querySelector('.CodeMirror-dialog');
    return (
      to.line < 3 &&
      dialog &&
      dialog.getBoundingClientRect().bottom - 4 > cm.cursorCoords(to, 'window').top
    );
  }

  function isFindCommand(cmd) {
    return ['findNext', 'findPrev', 'findPersistentNext', 'findPersistentPrev'].includes(cmd);
  }

  function isSearchCommand(cmd) {
    return ['find', 'findPersistent'].includes(cmd);
  }


  function findNext(cm, rev, callback) {
    cm.operation(function () {
      let state = getSearchState(cm)
      let cursor = getSearchCursor(
        cm,
        state.query,
        rev ? state.posFrom : state.posTo,
      )
      if (!cursor.find(rev)) {
        cursor = getSearchCursor(
          cm,
          state.query,
          rev
            ? CodeMirror.Pos(cm.lastLine())
            : CodeMirror.Pos(cm.firstLine(), 0),
        )
        if (!cursor.find(rev)) return
      }
      cm.setSelection(cursor.from(), cursor.to())
      cm.scrollIntoView({ from: cursor.from(), to: cursor.to() }, 20)
      state.posFrom = cursor.from()
      state.posTo = cursor.to()
      if (callback) callback(cursor.from(), cursor.to())
    })
  }

  function clearSearch(cm) {
    cm.operation(function () {
      let state = getSearchState(cm)
      state.lastQuery = state.query
      if (!state.query) return
      state.query = state.queryText = null
      cm.removeOverlay(state.overlay)
      if (state.annotate) {
        state.annotate.clear()
        state.annotate = null
      }
    })
  }

  function el(tag, attrs) {
    let element = tag
      ? document.createElement(tag)
      : document.createDocumentFragment()
    for (let key in attrs) {
      element[key] = attrs[key]
    }
    for (let i = 2; i < arguments.length; i++) {
      let child = arguments[i]
      element.appendChild(
        typeof child == 'string' ? document.createTextNode(child) : child,
      )
    }
    return element
  }

  function getQueryDialog(cm) {
    let label = el(
      'label',
      { className: 'CodeMirror-search-label' },
      cm.phrase('Search:'),
      el('input', {
        type: 'text',
        style: 'width: 10em',
        className: 'CodeMirror-search-field',
        id: 'CodeMirror-search-field',
      }),
    )
    label.setAttribute('for', 'CodeMirror-search-field')
    return el(
      '',
      null,
      label,
      ' ',
      el(
        'span',
        { style: 'color: #666', className: 'CodeMirror-search-hint' },
        cm.phrase('(Use /re/ syntax for regexp search)'),
      ),
    )
  }
  function getReplaceQueryDialog(cm) {
    return el(
      '',
      null,
      ' ',
      el('input', {
        type: 'text',
        style: 'width: 10em',
        className: 'CodeMirror-search-field',
      }),
      ' ',
      el(
        'span',
        { style: 'color: #666', className: 'CodeMirror-search-hint' },
        cm.phrase('(Use /re/ syntax for regexp search)'),
      ),
    )
  }
  function getReplacementQueryDialog(cm) {
    return el(
      '',
      null,
      el('span', { className: 'CodeMirror-search-label' }, cm.phrase('With:')),
      ' ',
      el('input', {
        type: 'text',
        style: 'width: 10em',
        className: 'CodeMirror-search-field',
      }),
    )
  }
  function getDoReplaceConfirm(cm) {
    return el(
      '',
      null,
      el(
        'span',
        { className: 'CodeMirror-search-label' },
        cm.phrase('Replace?'),
      ),
      ' ',
      el('button', {}, cm.phrase('Yes')),
      ' ',
      el('button', {}, cm.phrase('No')),
      ' ',
      el('button', {}, cm.phrase('All')),
      ' ',
      el('button', {}, cm.phrase('Stop')),
    )
  }

  function replaceAll(cm, query, text) {
    cm.operation(function () {
      for (let cursor = getSearchCursor(cm, query); cursor.findNext();) {
        if (typeof query != 'string') {
          let match = cm.getRange(cursor.from(), cursor.to()).match(query)
          cursor.replace(
            text.replace(/\$(\d)/g, function (_, i) {
              return match[i]
            }),
          )
        } else cursor.replace(text)
      }
    })
  }

  function replace(cm, all) {
    if (cm.getOption('readOnly')) return;

    const query = cm.getSelection() || getSearchState(cm).lastQuery;
    const dialogText = all ? cm.phrase('Replace all:') : cm.phrase('Replace:');

    const fragment = createReplaceDialog(cm, dialogText, query);
    dialog(cm, fragment, dialogText, query, function (query) {
      if (!query) return;
      handleQueryReplacement(cm, all, query);
    });
  }

  function createReplaceDialog(cm, dialogText, query) {
    return el(
      '',
      null,
      el('span', { className: 'CodeMirror-search-label' }, dialogText),
      getReplaceQueryDialog(cm),
    );
  }

  function handleQueryReplacement(cm, all, query) {
    query = parseQuery(query);

    if (all) {
      replaceAll(cm, query, getReplacementText());
    } else {
      clearSearch(cm);
      let cursor = getSearchCursor(cm, query, cm.getCursor('from'));
      processReplaceQuery(cm, cursor, query);
    }
  }

  function processReplaceQuery(cm, cursor, query) {
    const advance = () => {
      let start = cursor.from();
      let match = cursor.findNext();

      if (!match) {
        cursor = getSearchCursor(cm, query);
        match = cursor.findNext();

        if (!match || (start && cursor.from().line === start.line && cursor.from().ch === start.ch)) {
          return;
        }
      }


      cm.setSelection(cursor.from(), cursor.to());
      cm.scrollIntoView({ from: cursor.from(), to: cursor.to() });
      showReplaceConfirmDialog(cm, match);
    };

    advance();
  }

  function getReplacementText() {
    return parseString(getReplacementQueryDialog(cm));
  }

  function showReplaceConfirmDialog(cm, match) {
    confirmDialog(
      cm,
      getDoReplaceConfirm(cm),
      cm.phrase('Replace?'),
      [
        function () {
          doReplace(match);
        },
        advance,
        function () {
          replaceAll(cm, query, getReplacementText());
        },
      ]
    );
  }

  function doReplace(match) {
    const cursor = getSearchCursor(cm, query);
    cursor.replace(
      typeof query == 'string'
        ? getReplacementText()
        : getReplacementText().replace(/\$(\d)/g, function (_, i) {
          return match[i];
        })
    );
  }


  CodeMirror.commands.find = function (cm) {
    clearSearch(cm)
    doSearch(cm)
  }
  CodeMirror.commands.findPersistent = function (cm) {
    clearSearch(cm)
    doSearch(cm, false, true)
  }
  CodeMirror.commands.findPersistentNext = function (cm) {
    doSearch(cm, false, true, true)
  }
  CodeMirror.commands.findPersistentPrev = function (cm) {
    doSearch(cm, true, true, true)
  }
  CodeMirror.commands.findNext = doSearch
  CodeMirror.commands.findPrev = function (cm) {
    doSearch(cm, true)
  }
  CodeMirror.commands.clearSearch = clearSearch
  CodeMirror.commands.replace = replace
  CodeMirror.commands.replaceAll = function (cm) {
    replace(cm, true)
  }
})
